""" Class for representing individuals within the `lineage` framework. """

"""
Copyright (C) 2016 Andrew Riha

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
import re

import numpy as np
import pandas as pd

import lineage
from lineage.snps import (SNPs, get_assembly, get_chromosomes, get_chromosomes_summary,
                          get_snp_count, sort_snps, determine_sex)


class Individual(object):
    """ Object used to represent and interact with an individual.

    The ``Individual`` object maintains information about an individual. The object provides
    methods for loading an individual's genetic data (SNPs) and normalizing it for use with the
    `lineage` framework.

    """

    def __init__(self, name, raw_data=None, output_dir='output'):
        """ Initialize an ``Individual`` object.

        Parameters
        ----------
        name : str
            name of the individual
        raw_data : list or str
            path(s) to file(s) with raw genotype data
        output_dir : str
            path to output directory
        """
        self._name = name
        self._output_dir = output_dir
        self._snps = None
        self._build = None
        self._source = []
        self._discrepant_positions_file_count = 0
        self._discrepant_genotypes_file_count = 0
        self._discrepant_positions = pd.DataFrame()
        self._discrepant_genotypes = pd.DataFrame()

        if raw_data is not None:
            self.load_snps(raw_data)

    def __repr__(self):
        return 'Individual({!r})'.format(self._name)

    @property
    def name(self):
        """ Get this ``Individual``'s name.

        Returns
        -------
        str
        """
        return self._name

    @property
    def snps(self):
        """ Get a copy of this ``Individual``'s SNPs.

        Returns
        -------
        pandas.DataFrame
        """
        if self._snps is not None:
            return self._snps.copy()
        else:
            return None

    @property
    def snp_count(self):
        """ Count of SNPs loaded for this ``Individual``.

        Returns
        -------
        int
        """
        return get_snp_count(self._snps)

    @property
    def chromosomes(self):
        """ Get the chromosomes of this ``Individual``'s SNPs.

        Returns
        -------
        list
            list of str chromosomes (e.g., ['1', '2', '3', 'MT'], empty list if no chromosomes
        """
        return get_chromosomes(self._snps)

    @property
    def chromosomes_summary(self):
        """ Summary of the chromosomes represented by this ``Individual``'s SNPs.

        Returns
        -------
        str
            human-readable listing of chromosomes (e.g., '1-3, MT'), empty str if no chromosomes
        """
        return get_chromosomes_summary(self._snps)

    @property
    def build(self):
        """ Get the build of this ``Individual``'s SNPs.

        Returns
        -------
        int
        """
        return self._build

    @property
    def assembly(self):
        """ Get the assembly of this ``Individual``'s SNPs.

        Returns
        -------
        str
        """
        return get_assembly(self._build)

    @property
    def source(self):
        """ Summary of the SNP data source(s) for this ``Individual``'s SNPs.

        Returns
        -------
        str
        """
        return ', '.join(self._source)

    @property
    def sex(self):
        """ Sex of this ``Individual`` derived from this ``Individual``'s SNPs.

        Returns
        -------
        str
            'Male' or 'Female' if detected, else empty str
        """
        return determine_sex(self._snps)

    @property
    def discrepant_positions(self):
        """ SNPs with discrepant positions discovered while loading SNPs.

        Returns
        -------
        pandas.DataFrame
        """
        return self._discrepant_positions

    @property
    def discrepant_genotypes(self):
        """ SNPs with discrepant genotypes discovered while loading SNPs.

        Returns
        -------
        pandas.DataFrame
        """
        return self._discrepant_genotypes

    @property
    def discrepant_snps(self):
        """ SNPs with discrepant positions and / or genotypes discovered while loading SNPs.

        Returns
        -------
        pandas.DataFrame
        """
        df = self._discrepant_positions.append(self._discrepant_genotypes)
        if len(df) > 1:
            df = df.drop_duplicates()
        return df

    def load_snps(self, raw_data, discrepant_snp_positions_threshold=100,
                  discrepant_genotypes_threshold=500, save_output=False):
        """ Load raw genotype data.

        Parameters
        ----------
        raw_data : list or str
            path(s) to file(s) with raw genotype data
        discrepant_snp_positions_threshold : int
            threshold for discrepant SNP positions between existing data and data to be loaded,
            a large value could indicate mismatched genome assemblies
        discrepant_genotypes_threshold : int
            threshold for discrepant genotype data between existing data and data to be loaded,
            a large value could indicated mismatched individuals
        save_output : bool
            specifies whether to save discrepant SNP output to CSV files in the output directory
        """
        if type(raw_data) is list:
            for file in raw_data:
                self._load_snps_helper(file, discrepant_snp_positions_threshold,
                                       discrepant_genotypes_threshold, save_output)
        elif type(raw_data) is str:
            self._load_snps_helper(raw_data, discrepant_snp_positions_threshold,
                                   discrepant_genotypes_threshold, save_output)
        else:
            raise TypeError('invalid filetype')

    def _load_snps_helper(self, file, discrepant_snp_positions_threshold,
                          discrepant_genotypes_threshold, save_output):
        print('Loading ' + os.path.relpath(file))
        discrepant_positions, discrepant_genotypes = \
            self._add_snps(SNPs(file), discrepant_snp_positions_threshold,
                           discrepant_genotypes_threshold, save_output)

        self._discrepant_positions = self._discrepant_positions.append(discrepant_positions,
                                                                       sort=True)
        self._discrepant_genotypes = self._discrepant_genotypes.append(discrepant_genotypes,
                                                                       sort=True)

    def save_snps(self, filename=None):
        """ Save SNPs to file.

        Parameters
        ----------
        filename : str
            filename for file to save

        Returns
        -------
        str
            path to file in output directory if SNPs were saved, else empty str
        """
        comment = '# Source(s): {}\n'\
                  '# Assembly: {}\n' \
                  '# SNPs: {}\n' \
                  '# Chromosomes: {}\n'.format(self.source,
                                               self.assembly,
                                               self.snp_count,
                                               self.chromosomes_summary)

        if filename is None:
            filename = self.get_var_name() + '_lineage_' + self.assembly + '.csv'

        return lineage.save_df_as_csv(self._snps, self._output_dir, filename, comment=comment,
                                      header=['chromosome', 'position', 'genotype'])

    def save_discrepant_positions(self, filename=None):
        """ Save SNPs with discrepant positions to file.

        Parameters
        ----------
        filename : str
            filename for file to save

        Returns
        -------
        str
            path to file in output directory if SNPs were saved, else empty str
        """
        return self._save_discrepant_snps_file(self.discrepant_positions, 'discrepant_positions',
                                               filename)

    def save_discrepant_genotypes(self, filename=None):
        """ Save SNPs with discrepant genotypes to file.

        Parameters
        ----------
        filename : str
            filename for file to save

        Returns
        -------
        str
            path to file in output directory if SNPs were saved, else empty str
        """
        return self._save_discrepant_snps_file(self.discrepant_genotypes, 'discrepant_genotypes',
                                               filename)

    def save_discrepant_snps(self, filename=None):
        """ Save SNPs with discrepant positions and / or genotypes to file.

        Parameters
        ----------
        filename : str
            filename for file to save

        Returns
        -------
        str
            path to file in output directory if SNPs were saved, else empty str
        """
        return self._save_discrepant_snps_file(self.discrepant_snps, 'discrepant_snps', filename)

    def _save_discrepant_snps_file(self, df, name, filename):
        if filename is None:
            filename = self.get_var_name() + '_' + name + '.csv'

        return lineage.save_df_as_csv(df, self._output_dir, filename,
                                      comment='# Source(s): {}\n'.format(self.source))

    def get_var_name(self):
        return self._clean_string(self.name)

    def _clean_string(self, s):
        """ Clean a string so that it can be a valid Python variable
        name.

        Parameters
        ----------
        s : str
            string to clean

        Returns
        -------
        str
            cleaned string that can be used as a variable name
        """
        # http://stackoverflow.com/a/3305731
        return re.sub('\W|^(?=\d)', '_', s)

    def remap_snps(self, target_assembly, complement_bases=True):
        """ Remap the SNP coordinates of this ``Individual`` from one assembly to another.

        This method is a wrapper for `remap_snps` in the ``Lineage`` class.

        This method uses the assembly map endpoint of the Ensembl REST API service to convert SNP
        coordinates / positions from one assembly to another. After remapping, the coordinates /
        positions for the ``Individual``'s SNPs will be that of the target assembly.

        If the SNPs are already mapped relative to the target assembly, remapping will not be
        performed.

        Parameters
        ----------
        target_assembly : {'NCBI36', 'GRCh37', 'GRCh38', 36, 37, 38}
            assembly to remap to
        complement_bases : bool
            complement bases when remapping SNPs to the minus strand

        Returns
        -------
        chromosomes_remapped : list of str
            chromosomes remapped; empty if None
        chromosomes_not_remapped : list of str
            chromosomes not remapped; empty if None

        Notes
        -----
        An assembly is also know as a "build." For example:

        Assembly NCBI36 = Build 36
        Assembly GRCh37 = Build 37
        Assembly GRCh38 = Build 38

        See https://www.ncbi.nlm.nih.gov/assembly for more information about assemblies and
        remapping.

        References
        ----------
        ..[1] Ensembl, Assembly Map Endpoint,
          http://rest.ensembl.org/documentation/info/assembly_map
        """
        from lineage import Lineage
        l = Lineage()
        return l.remap_snps(self, target_assembly, complement_bases)

    def _set_snps(self, snps, build=37):
        """ Set `_snps` and `_build` properties of this ``Individual``.

        Notes
        -----
        Intended to be used internally to `lineage`.

        Parameters
        ----------
        snps : pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        build : int
            build of this ``Individual``'s SNPs
        """
        self._snps = snps
        self._build = build

    def _add_snps(self, snps, discrepant_snp_positions_threshold,
                  discrepant_genotypes_threshold, save_output):
        """ Add SNPs to this Individual.

        Parameters
        ----------
        snps : SNPs
            SNPs to add
        discrepant_snp_positions_threshold : int
            see above
        discrepant_genotypes_threshold : int
            see above
        save_output
            see above

        Returns
        -------
        discrepant_positions : pandas.DataFrame
        discrepant_genotypes : pandas.DataFrame
        """
        discrepant_positions = pd.DataFrame()
        discrepant_genotypes = pd.DataFrame()

        if snps.snps is None:
            return discrepant_positions, discrepant_genotypes

        build = snps.build
        source = [s.strip() for s in snps.source.split(',')]

        if not snps.build_detected:
            print('build not detected, assuming build {}'.format(snps.build))

        if self._build is None:
            self._build = build
        elif self._build != build:
            print('build / assembly mismatch between current build of SNPs and SNPs being loaded')

        # ensure there area always two X alleles
        snps = self._double_single_alleles(snps.snps, 'X')

        if self._snps is None:
            self._source.extend(source)
            self._snps = snps
        else:
            common_snps = self._snps.join(snps, how='inner', rsuffix='_added')

            discrepant_positions = common_snps.loc[
                (common_snps['chrom'] != common_snps['chrom_added']) |
                (common_snps['pos'] != common_snps['pos_added'])]

            if 0 < len(discrepant_positions) < discrepant_snp_positions_threshold:
                print(str(len(discrepant_positions)) + ' SNP positions were discrepant; '
                      'keeping original positions')

                if save_output:
                    self._discrepant_positions_file_count += 1
                    lineage.save_df_as_csv(discrepant_positions, self._output_dir,
                                           self.get_var_name() + '_discrepant_positions_' + str(
                                               self._discrepant_positions_file_count) + '.csv')
            elif len(discrepant_positions) >= discrepant_snp_positions_threshold:
                print('too many SNPs differ in position; ensure same genome build is being used')
                return discrepant_positions, discrepant_genotypes

            # remove null genotypes
            common_snps = common_snps.loc[~common_snps['genotype'].isnull() &
                                          ~common_snps['genotype_added'].isnull()]

            # discrepant genotypes are where alleles are not equivalent (i.e., alleles are not the
            # same and not swapped)
            discrepant_genotypes = common_snps.loc[
                ((common_snps['genotype'].str.len() == 1) &
                 (common_snps['genotype_added'].str.len() == 1) &
                 ~(common_snps['genotype'].str[0] == common_snps['genotype_added'].str[0])) |
                ((common_snps['genotype'].str.len() == 2) &
                 (common_snps['genotype_added'].str.len() == 2) &
                 ~((common_snps['genotype'].str[0] == common_snps['genotype_added'].str[0]) &
                   (common_snps['genotype'].str[1] == common_snps['genotype_added'].str[1])) &
                 ~((common_snps['genotype'].str[0] == common_snps['genotype_added'].str[1]) &
                   (common_snps['genotype'].str[1] == common_snps['genotype_added'].str[0])))]

            if 0 < len(discrepant_genotypes) < discrepant_genotypes_threshold:
                print(str(len(discrepant_genotypes)) + ' SNP genotypes were discrepant; '
                      'marking those as null')

                if save_output:
                    self._discrepant_genotypes_file_count += 1
                    lineage.save_df_as_csv(discrepant_genotypes, self._output_dir,
                                           self.get_var_name() + '_discrepant_genotypes_' + str(
                                               self._discrepant_genotypes_file_count) + '.csv')
            elif len(discrepant_genotypes) >= discrepant_genotypes_threshold:
                print('too many SNPs differ in their genotype; ensure file is for same '
                      'individual')
                return discrepant_positions, discrepant_genotypes

            # add new SNPs
            self._source.extend(source)
            self._snps = self._snps.combine_first(snps)
            self._snps.loc[discrepant_genotypes.index, 'genotype'] = np.nan

            # combine_first converts position to float64, so convert it back to int64
            self._snps['pos'] = self._snps['pos'].astype(np.int64)

        self._snps = sort_snps(self._snps)

        return discrepant_positions, discrepant_genotypes

    @staticmethod
    def _double_single_alleles(df, chrom):
        """ Double any single alleles in the specified chromosome.

        Parameters
        ----------
        df : pandas.DataFrame
            SNPs
        chrom : str
            chromosome of alleles to double

        Returns
        -------
        df : pandas.DataFrame
            SNPs with specified chromosome's single alleles doubled
        """
        # find all single alleles of the specified chromosome
        single_alleles = np.where((df['chrom'] == chrom) &
                                  (df['genotype'].str.len() == 1))[0]

        # double those alleles
        df.ix[single_alleles, 'genotype'] = df.ix[single_alleles, 'genotype'] * 2

        return df



""" Ensembl REST client.

Notes
-----
Modified from https://github.com/Ensembl/ensembl-rest/wiki/Example-Python-Client.

References
----------
..[1] Yates et. al. (doi:10.1093/bioinformatics/btu613),
  http://europepmc.org/search/?query=DOI:10.1093/bioinformatics/btu613
..[2] Zerbino et. al. (doi.org/10.1093/nar/gkx1098), https://doi.org/10.1093/nar/gkx1098

"""

"""
Copyright [1999-2015] Wellcome Trust Sanger Institute and the EMBL-European Bioinformatics Institute
Copyright [2016-2017] EMBL-European Bioinformatics Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and

"""

"""
Copyright (C) 2017 Andrew Riha

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import json
import sys
import time
import urllib.error
import urllib.parse
import urllib.request


class EnsemblRestClient(object):
    def __init__(self, server='http://rest.ensembl.org', reqs_per_sec=15):
        self.server = server
        self.reqs_per_sec = reqs_per_sec
        self.req_count = 0
        self.last_req = 0

    def perform_rest_action(self, endpoint, hdrs=None, params=None):
        if hdrs is None:
            hdrs = {}

        if 'Content-Type' not in hdrs:
            hdrs['Content-Type'] = 'application/json'

        if params:
            endpoint += '?' + urllib.parse.urlencode(params)

        data = None

        # check if we need to rate limit ourselves
        if self.req_count >= self.reqs_per_sec:
            delta = time.time() - self.last_req
            if delta < 1:
                time.sleep(1 - delta)
            self.last_req = time.time()
            self.req_count = 0

        try:
            request = urllib.request.Request(self.server + endpoint, headers=hdrs)

            with urllib.request.urlopen(request) as response:
                content = response.read().decode('utf-8')

            if content:
                data = json.loads(content)
            self.req_count += 1

        except urllib.error.HTTPError as e:
            # check if we are being rate limited by the server
            if e.code == 429:
                if 'Retry-After' in e.headers:
                    retry = e.headers['Retry-After']
                    time.sleep(float(retry))
                    self.perform_rest_action(endpoint, hdrs, params)
            else:
                sys.stderr.write(
                    'Request failed for {0}: Status code: {1.code} Reason: {1.reason}\n'.format(
                        endpoint, e))

        return data



""" lineage

tools for genetic genealogy and the analysis of consumer DNA test results

"""

"""
Copyright (C) 2016 Andrew Riha

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime
import os
import numpy as np
import pandas as pd

# http://mikegrouchy.com/blog/2012/05/be-pythonic-__init__py.html
from lineage.ensembl import EnsemblRestClient
from lineage.individual import Individual
from lineage.resources import Resources
from lineage.snps import sort_snps
from lineage.visualization import plot_chromosomes

# set version string with Versioneer
from lineage._version import get_versions
__version__ = get_versions()['version']
del get_versions


class Lineage(object):
    """ Object used to interact with the `lineage` framework. """

    def __init__(self, output_dir='output', resources_dir='resources'):
        """ Initialize a ``Lineage`` object.

        Parameters
        ----------
        output_dir : str
            name / path of output directory
        resources_dir
            name / path of resources directory
        """
        self._output_dir = os.path.abspath(output_dir)
        self._ensembl_rest_client = EnsemblRestClient()
        self._resources = Resources(resources_dir=resources_dir,
                                    ensembl_rest_client=self._ensembl_rest_client)

    def create_individual(self, name, raw_data=None):
        """ Initialize an individual in the context of the `lineage` framework.

        Parameters
        ----------
        name : str
            name of the individual
        raw_data : list or str
            path(s) to file(s) with raw genotype data

        Returns
        -------
        Individual
            ``Individual`` initialized in the context of the `lineage` framework
        """
        return Individual(name, raw_data, self._output_dir)

    def download_example_datasets(self):
        """ Download example datasets from `openSNP <https://opensnp.org>`_.

        Per openSNP, "the data is donated into the public domain using `CC0 1.0
        <http://creativecommons.org/publicdomain/zero/1.0/>`_."

        Returns
        -------
        paths : list of str or None
            paths to example datasets

        References
        ----------
        ..[1] Greshake B, Bayer PE, Rausch H, Reda J (2014), "openSNP-A Crowdsourced Web Resource
          for Personal Genomics," PLOS ONE, 9(3): e89204,
          https://doi.org/10.1371/journal.pone.0089204
        """
        paths = self._resources.download_example_datasets()

        if None in paths:
            print('Example dataset(s) not currently available')

        return paths

    def remap_snps(self, individual, target_assembly, complement_bases=True):
        """ Remap the SNP coordinates of an individual from one assembly to another.

        This method uses the assembly map endpoint of the Ensembl REST API service (via
        ``Resources``'s ``EnsemblRestClient``) to convert SNP coordinates / positions from one
        assembly to another. After remapping, the coordinates / positions for the individual's
        SNPs will be that of the target assembly.

        If the SNPs are already mapped relative to the target assembly, remapping will not be
        performed.

        Parameters
        ----------
        target_assembly : {'NCBI36', 'GRCh37', 'GRCh38', 36, 37, 38}
            assembly to remap to
        complement_bases : bool
            complement bases when remapping SNPs to the minus strand

        Returns
        -------
        chromosomes_remapped : list of str
            chromosomes remapped; empty if None
        chromosomes_not_remapped : list of str
            chromosomes not remapped; empty if None

        Notes
        -----
        An assembly is also know as a "build." For example:

        Assembly NCBI36 = Build 36
        Assembly GRCh37 = Build 37
        Assembly GRCh38 = Build 38

        See https://www.ncbi.nlm.nih.gov/assembly for more information about assemblies and
        remapping.

        References
        ----------
        ..[1] Ensembl, Assembly Map Endpoint,
          http://rest.ensembl.org/documentation/info/assembly_map
        """
        chromosomes_remapped = []
        chromosomes_not_remapped = []

        snps = individual.snps

        if snps is None:
            print('No SNPs to remap')
            return chromosomes_remapped, chromosomes_not_remapped
        else:
            chromosomes_not_remapped = list(snps['chrom'].unique())

        valid_assemblies = ['NCBI36', 'GRCh37', 'GRCh38', 36, 37, 38]

        if target_assembly not in valid_assemblies:
            print('Invalid target assembly')
            return chromosomes_remapped, chromosomes_not_remapped

        if isinstance(target_assembly, int):
            if target_assembly == 36:
                target_assembly = 'NCBI36'
            else:
                target_assembly = 'GRCh' + str(target_assembly)

        if individual.build == 36:
            source_assembly = 'NCBI36'
        else:
            source_assembly = 'GRCh' + str(individual.build)

        if source_assembly == target_assembly:
            return chromosomes_remapped, chromosomes_not_remapped

        assembly_mapping_data = \
            self._resources.get_assembly_mapping_data(source_assembly, target_assembly)

        if assembly_mapping_data is None:
            return chromosomes_remapped, chromosomes_not_remapped

        for chrom in snps['chrom'].unique():
            # extract SNPs for this chrom for faster remapping
            temp = pd.DataFrame(snps.loc[snps['chrom'] == chrom])

            temp['remapped'] = False

            if chrom in assembly_mapping_data:
                chromosomes_remapped.append(chrom)
                chromosomes_not_remapped.remove(chrom)
                mappings = assembly_mapping_data[chrom]
            else:
                print('Chromosome ' + chrom + ' not remapped; '
                      'removing chromosome from SNPs for consistency')
                snps = snps.drop(snps.loc[snps['chrom'] == chrom].index)
                continue

            pos_start = int(temp['pos'].describe()['min'])
            pos_end = int(temp['pos'].describe()['max'])

            for mapping in mappings['mappings']:
                # skip if mapping is outside of range of SNP positions
                if mapping['original']['end'] <= pos_start or \
                        mapping['original']['start'] >= pos_end:
                    continue

                orig_range_len = mapping['original']['end'] - mapping['original']['start']
                mapped_range_len = mapping['mapped']['end'] - mapping['mapped']['start']

                orig_region = mapping['original']['seq_region_name']
                mapped_region = mapping['mapped']['seq_region_name']

                if orig_region != mapped_region:
                    print('discrepant chroms')
                    continue

                if orig_range_len != mapped_range_len:
                    print('discrepant coords')  # observed when mapping NCBI36 -> GRCh38
                    continue

                # find the SNPs that are being remapped for this mapping
                snp_indices = temp.loc[~temp['remapped'] &
                                   (temp['pos'] >= mapping['original']['start']) &
                                   (temp['pos'] <= mapping['original']['end'])].index

                if len(snp_indices) > 0:
                    # remap the SNPs
                    if mapping['mapped']['strand'] == -1:
                        # flip and (optionally) complement since we're mapping to minus strand
                        diff_from_start = temp.loc[snp_indices, 'pos'] - mapping['original']['start']
                        temp.loc[snp_indices, 'pos'] = mapping['mapped']['end'] - diff_from_start

                        if complement_bases:
                            snps.loc[snp_indices, 'genotype'] = \
                                temp.loc[snp_indices, 'genotype'].apply(self._complement_bases)
                    else:
                        # mapping is on same (plus) strand, so just remap based on offset
                        offset = mapping['mapped']['start'] - mapping['original']['start']
                        temp.loc[snp_indices, 'pos'] = temp['pos'] + offset

                    # mark these SNPs as remapped
                    temp.loc[snp_indices, 'remapped'] = True

            # update SNP positions for this chrom
            snps.loc[temp.index, 'pos'] = temp['pos']

        individual._set_snps(sort_snps(snps), int(target_assembly[-2:]))

        return chromosomes_remapped, chromosomes_not_remapped

    def _complement_bases(self, genotype):
        if pd.isnull(genotype):
            return np.nan

        complement = ''

        for base in list(genotype):
            if base == 'A':
                complement += 'T'
            elif base == 'G':
                complement += 'C'
            elif base == 'C':
                complement += 'G'
            elif base == 'T':
                complement += 'A'
            else:
                complement += base

        return complement

    def find_discordant_snps(self, individual1, individual2, individual3=None, save_output=False):
        """ Find discordant SNPs between two or three individuals.

        Parameters
        ----------
        individual1 : Individual
            reference individual (child if `individual2` and `individual3` are parents)
        individual2 : Individual
            comparison individual
        individual3 : Individual
            other parent if `individual1` is child and `individual2` is a parent
        save_output : bool
            specifies whether to save output to a CSV file in the output directory

        Returns
        -------
        pandas.DataFrame
            discordant SNPs and associated genetic data

        References
        ----------
        ..[1] David Pike, "Search for Discordant SNPs in Parent-Child
          Raw Data Files," David Pike's Utilities,
          http://www.math.mun.ca/~dapike/FF23utils/pair-discord.php
        ..[2] David Pike, "Search for Discordant SNPs when given data
          for child and both parents," David Pike's Utilities,
          http://www.math.mun.ca/~dapike/FF23utils/trio-discord.php
        """
        self._remap_snps_to_GRCh37([individual1, individual2, individual3])

        df = individual1.snps

        # remove nulls for reference individual
        df = df.loc[df['genotype'].notnull()]

        # add SNPs shared with `individual2`
        df = df.join(individual2.snps['genotype'], rsuffix='2')

        genotype1 = 'genotype_' + individual1.get_var_name()
        genotype2 = 'genotype_' + individual2.get_var_name()

        if individual3 is None:
            df = df.rename(columns={'genotype': genotype1, 'genotype2': genotype2})

            # find discordant SNPs between reference and comparison individuals
            df = df.loc[df[genotype2].notnull() &
                        ((df[genotype1].str.len() == 1) &
                         (df[genotype2].str.len() == 1) &
                         (df[genotype1] != df[genotype2])) |
                        ((df[genotype1].str.len() == 2) &
                         (df[genotype2].str.len() == 2) &
                         (df[genotype1].str[0] != df[genotype2].str[0]) &
                         (df[genotype1].str[0] != df[genotype2].str[1]) &
                         (df[genotype1].str[1] != df[genotype2].str[0]) &
                         (df[genotype1].str[1] != df[genotype2].str[1]))]
            if save_output:
                save_df_as_csv(df, self._output_dir, 'discordant_snps_' +
                               individual1.get_var_name() + '_' +
                               individual2.get_var_name() + '_GRCh37.csv')
        else:
            # add SNPs shared with `individual3`
            df = df.join(individual3.snps['genotype'], rsuffix='3')

            genotype3 = 'genotype_' + individual3.get_var_name()

            df = df.rename(columns={'genotype': genotype1, 'genotype2': genotype2,
                                    'genotype3': genotype3})

            # find discordant SNPs between child and two parents
            df = df.loc[(df[genotype2].notnull() &
                         ((df[genotype1].str.len() == 1) &
                          (df[genotype2].str.len() == 1) &
                          (df[genotype1] != df[genotype2])) |
                         ((df[genotype1].str.len() == 2) &
                          (df[genotype2].str.len() == 2) &
                          (df[genotype1].str[0] != df[genotype2].str[0]) &
                          (df[genotype1].str[0] != df[genotype2].str[1]) &
                          (df[genotype1].str[1] != df[genotype2].str[0]) &
                          (df[genotype1].str[1] != df[genotype2].str[1]))) |
                        (df[genotype3].notnull() &
                         ((df[genotype1].str.len() == 1) &
                          (df[genotype3].str.len() == 1) &
                          (df[genotype1] != df[genotype3])) |
                         ((df[genotype1].str.len() == 2) &
                          (df[genotype3].str.len() == 2) &
                          (df[genotype1].str[0] != df[genotype3].str[0]) &
                          (df[genotype1].str[0] != df[genotype3].str[1]) &
                          (df[genotype1].str[1] != df[genotype3].str[0]) &
                          (df[genotype1].str[1] != df[genotype3].str[1]))) |
                        (df[genotype2].notnull() & df[genotype3].notnull() &
                         (df[genotype2].str.len() == 2) &
                         (df[genotype2].str[0] == df[genotype2].str[1]) &
                         (df[genotype2] == df[genotype3]) &
                         (df[genotype1] != df[genotype2]))]

            if save_output:
                save_df_as_csv(df, self._output_dir, 'discordant_snps_' +
                               individual1.get_var_name() + '_' +
                               individual2.get_var_name() + '_' +
                               individual3.get_var_name() + '_GRCh37.csv')

        return df

    def find_shared_dna(self, individual1, individual2, cM_threshold=0.75,
                        snp_threshold=1100, shared_genes=False, save_output=True):
        """ Find the shared DNA between two individuals.

        Computes the genetic distance in centiMorgans (cMs) between SNPs using the HapMap Phase II
        GRCh37 genetic map. Applies thresholds to determine the shared DNA. Plots shared DNA.
        Optionally determines shared genes (i.e., genes that are transcribed from the shared DNA).

        All output is saved to the output directory as `CSV` or `PNG` files.

        Parameters
        ----------
        individual1 : Individual
        individual2 : Individual
        cM_threshold : float
            minimum centiMorgans for each shared DNA segment
        snp_threshold : int
            minimum SNPs for each shared DNA segment
        shared_genes : bool
            determine shared genes
        save_output : bool
            specifies whether to save output files in the output directory

        Returns
        -------
        one_chrom_shared_dna : pandas.DataFrame
            segments of shared DNA on one chromosome
        two_chrom_shared_dna : pandas.DataFrame
            segments of shared DNA on two chromosomes
        one_chrom_shared_genes : pandas.DataFrame
            shared genes on one chromosome
        two_chrom_shared_genes : pandas.DataFrame
            shared genes on two chromosomes
        """
        one_chrom_shared_genes = pd.DataFrame()
        two_chrom_shared_genes = pd.DataFrame()

        self._remap_snps_to_GRCh37([individual1, individual2])

        df = individual1.snps

        df = df.join(individual2.snps['genotype'], rsuffix='2', how='inner')

        genotype1 = 'genotype_' + individual1.get_var_name()
        genotype2 = 'genotype_' + individual2.get_var_name()

        df = df.rename(columns={'genotype': genotype1, 'genotype2': genotype2})

        one_x_chrom = self._is_one_individual_male([individual1, individual2])

        # determine the genetic distance between each SNP using the HapMap Phase II genetic map
        genetic_map, df = self._compute_snp_distances(df)

        # determine where individuals share an allele on one chromosome
        df['one_chrom_match'] = np.where(
            df[genotype1].isnull() | df[genotype2].isnull() |
            (df[genotype1].str[0] == df[genotype2].str[0]) |
            (df[genotype1].str[0] == df[genotype2].str[1]) |
            (df[genotype1].str[1] == df[genotype2].str[0]) |
            (df[genotype1].str[1] == df[genotype2].str[1]), True, False)

        # determine where individuals share alleles on both chromosomes
        df['two_chrom_match'] = np.where(
            df[genotype1].isnull() | df[genotype2].isnull() |
            ((df[genotype1].str.len() == 2) &
             (df[genotype2].str.len() == 2) &
             ((df[genotype1] == df[genotype2]) |
              ((df[genotype1].str[0] == df[genotype2].str[1]) &
               (df[genotype1].str[1] == df[genotype2].str[0])))), True, False)

        # compute shared DNA between individuals
        one_chrom_shared_dna = self._compute_shared_dna(df, genetic_map, 'one_chrom_match',
                                                        cM_threshold, snp_threshold, one_x_chrom)

        two_chrom_shared_dna = self._compute_shared_dna(df, genetic_map, 'two_chrom_match',
                                                        cM_threshold, snp_threshold, one_x_chrom)

        cytobands = self._resources.get_cytoBand_hg19()

        # plot data
        if save_output:
            if create_dir(self._output_dir):
                plot_chromosomes(one_chrom_shared_dna, two_chrom_shared_dna, cytobands,
                                 os.path.join(self._output_dir, 'shared_dna_' +
                                              individual1.get_var_name() + '_' +
                                              individual2.get_var_name() + '.png'),
                                 individual1.name + ' / ' + individual2.name + ' shared DNA', 37)

        one_chrom_shared_dna = self._convert_shared_dna_list_to_df(one_chrom_shared_dna)

        if len(one_chrom_shared_dna) > 0:
            if save_output:
                self._save_shared_dna_csv_format(one_chrom_shared_dna, 'one',
                                                 individual1.get_var_name(),
                                                 individual2.get_var_name())
            if shared_genes:
                one_chrom_shared_genes = self._compute_shared_genes(one_chrom_shared_dna, 'one',
                                                                    individual1.get_var_name(),
                                                                    individual2.get_var_name(),
                                                                    save_output)

        two_chrom_shared_dna = self._convert_shared_dna_list_to_df(two_chrom_shared_dna)

        if len(two_chrom_shared_dna) > 0:
            if save_output:
                self._save_shared_dna_csv_format(two_chrom_shared_dna, 'two',
                                                 individual1.get_var_name(),
                                                 individual2.get_var_name())
            if shared_genes:
                two_chrom_shared_genes = self._compute_shared_genes(two_chrom_shared_dna, 'two',
                                                                    individual1.get_var_name(),
                                                                    individual2.get_var_name(),
                                                                    save_output)

        return one_chrom_shared_dna, two_chrom_shared_dna, \
               one_chrom_shared_genes, two_chrom_shared_genes

    def _convert_shared_dna_list_to_df(self, shared_dna):
        df = pd.DataFrame(shared_dna, columns=['chrom', 'start', 'end', 'cMs', 'snps'])
        df.index.name = 'segment'
        df.index = df.index + 1
        return df

    def _compute_shared_genes(self, shared_dna, type, individual1_name, individual2_name,
                              save_output):
        knownGenes = self._resources.get_knownGene_hg19()
        kgXref = self._resources.get_kgXref_hg19()

        # http://seqanswers.com/forums/showthread.php?t=22336
        df = knownGenes.join(kgXref)

        shared_genes_dfs = []
        shared_genes = pd.DataFrame()

        for shared_segment in shared_dna.itertuples():
            # determine genes transcribed from this shared DNA segment
            temp = df.loc[(df['chrom'] == shared_segment.chrom) &
                          (df['txStart'] >= shared_segment.start) &
                          (df['txEnd'] <= shared_segment.end)].copy()

            # select subset of columns
            temp = temp[['geneSymbol', 'chrom', 'strand', 'txStart', 'txEnd', 'refseq',
                         'proteinID', 'description']]

            if len(temp) > 0:
                shared_genes_dfs.append(temp)

        if len(shared_genes_dfs) > 0:
            shared_genes = pd.concat(shared_genes_dfs, sort=True)

            if save_output:
                if type == 'one':
                    chroms = 'one_chrom'
                else:
                    chroms = 'two_chroms'

                file = 'shared_genes_' + chroms + '_' + individual1_name + '_' + \
                       individual2_name + '_GRCh37.csv'

                save_df_as_csv(shared_genes, self._output_dir, file)

        return shared_genes

    def _save_shared_dna_csv_format(self, shared_dna, type, individual1_name, individual2_name):
        if type == 'one':
            chroms = 'one_chrom'
        else:
            chroms = 'two_chroms'

        file = 'shared_dna_' + chroms + '_' + individual1_name + '_' + individual2_name + \
               '_GRCh37.csv'

        save_df_as_csv(shared_dna, self._output_dir, file, float_format='%.2f')

    def _is_one_individual_male(self, individuals):
        for individual in individuals:
            if individual.sex == 'Male':
                return True
        return False

    def _compute_snp_distances(self, df):
        genetic_map = self._resources.get_genetic_map_HapMapII_GRCh37()

        for chrom in df['chrom'].unique():
            if chrom not in genetic_map.keys():
                continue

            # create a new dataframe from the positions for the current chromosome
            temp = pd.DataFrame(df.loc[(df['chrom'] == chrom)]['pos'].values, columns=['pos'])

            # merge genetic map for this chrom
            temp = temp.append(genetic_map[chrom], ignore_index=True, sort=True)

            # sort based on pos
            temp = temp.sort_values('pos')

            # fill recombination rates forward
            temp['rate'] = temp['rate'].fillna(method='ffill')

            # assume recombination rate of 0 for SNPs upstream of first defined rate
            temp['rate'] = temp['rate'].fillna(0)

            # get difference between positions
            pos_diffs = np.ediff1d(temp['pos'])

            # compute cMs between each pos based on probabilistic recombination rate
            # https://www.biostars.org/p/123539/
            cMs_match_segment = (temp['rate'] * np.r_[pos_diffs, 0] / 1e6).values

            # add back into temp
            temp['cMs'] = np.r_[0, cMs_match_segment][:-1]

            temp = temp.reset_index()
            del temp['index']

            # use null `map` values to find locations of SNPs
            snp_indices = temp.loc[temp['map'].isnull()].index

            # use SNP indices to determine boundaries over which to sum cMs
            start_snp_ix = snp_indices + 1
            end_snp_ix = np.r_[snp_indices, snp_indices[-1]][1:] + 1
            snp_boundaries = np.c_[start_snp_ix, end_snp_ix]

            # sum cMs between SNPs to get total cM distance between SNPs
            # http://stackoverflow.com/a/7471967
            c = np.r_[0, temp['cMs'].cumsum()][snp_boundaries]
            cM_from_prev_snp = c[:, 1] - c[:, 0]

            # debug
            # temp.loc[snp_indices, 'cM_from_prev_snp'] = np.r_[0, cM_from_prev_snp][:-1]
            # temp.to_csv('debug.csv')

            # add back into df
            df.loc[(df['chrom'] == chrom), 'cM_from_prev_snp'] = np.r_[0, cM_from_prev_snp][:-1]

        return genetic_map, df

    def _compute_shared_dna(self, df, genetic_map, col, cM_threshold, snp_threshold, one_x_chrom):
        shared_dna = []

        for chrom in df['chrom'].unique():
            if chrom not in genetic_map.keys():
                continue

            # set two_chrom_match in non-PAR region to False if an individual is male
            if chrom == 'X' and col == 'two_chrom_match' and one_x_chrom:
                # https://www.ncbi.nlm.nih.gov/grc/human
                df.loc[(df['chrom'] == 'X') & (df['pos'] > 2699520) & (df['pos'] < 154931044),
                       'two_chrom_match'] = False

            # get consecutive strings of trues
            # http://stackoverflow.com/a/17151327
            a = df.loc[(df['chrom'] == chrom)][col].values
            a = np.r_[a, False]
            a_rshifted = np.roll(a, 1)
            starts = a & ~a_rshifted
            ends = ~a & a_rshifted
            a_starts = np.nonzero(starts)[0]
            a_starts = np.reshape(a_starts, (len(a_starts), 1))
            a_ends = np.nonzero(ends)[0]
            a_ends = np.reshape(a_ends, (len(a_ends), 1))

            matches = np.hstack((a_starts, a_ends))

            c = np.r_[0, df.loc[(df['chrom'] == chrom)]['cM_from_prev_snp'].cumsum()][matches]
            cMs_match_segment = c[:, 1] - c[:, 0]

            # get matching segments where total cMs is greater than the threshold
            matches_passed = matches[np.where(cMs_match_segment > cM_threshold)]

            # get indices where a_end boundary is adjacent to a_start, perhaps indicating a
            # discrepant SNP
            adjacent_ix = np.where(np.roll(matches_passed[:, 0], -1) - matches_passed[:, 1] == 1)

            # if there are adjacent segments
            if len(adjacent_ix[0]) != 0:
                matches_stitched = np.array([0, 0])
                prev = -1
                counter = 0

                # stitch together adjacent segments
                for x in matches_passed:
                    if prev != -1 and prev + 1 == x[0]:
                        matches_stitched[counter, 1] = x[1]
                    else:
                        matches_stitched = np.vstack((matches_stitched, x))
                        counter += 1

                    prev = x[1]

                matches_passed = matches_stitched[1:]

            snp_counts = matches_passed[:, 1] - matches_passed[:, 0]

            # apply SNP count threshold for each matching segment; we now have the shared DNA
            # segments that pass the centiMorgan and SNP thresholds
            matches_passed = matches_passed[np.where(snp_counts > snp_threshold)]

            # compute total cMs for each match segment
            c = np.r_[0, df.loc[(df['chrom'] == chrom)]['cM_from_prev_snp'].cumsum()][matches_passed]
            cMs_match_segment = c[:, 1] - c[:, 0]

            counter = 0
            # save matches for this chromosome
            if col == 'one_chrom_match':
                chrom_stain = 'one_chrom'
            else:
                chrom_stain = 'two_chrom'

            for x in matches_passed:
                shared_dna.append({'chrom': chrom,
                                   'start': df.loc[(df['chrom'] == chrom)].ix[x[0]].pos,
                                   'end': df.loc[(df['chrom'] == chrom)].ix[x[1] - 1].pos,
                                   'cMs': cMs_match_segment[counter],
                                   'snps': x[1] - x[0],
                                   'gie_stain': chrom_stain})
                counter += 1
        return shared_dna

    @staticmethod
    def _remap_snps_to_GRCh37(individuals):
        for i in individuals:
            if i is None:
                continue

            i.remap_snps(37)


def create_dir(path):
    """ Create directory specified by `path` if it doesn't already exist.

    Parameters
    ----------
    path : str
        path to directory

    Returns
    -------
    bool
        True if `path` exists
    """
    # https://stackoverflow.com/a/5032238
    try:
        os.makedirs(path, exist_ok=True)
    except Exception as err:
        print(err)
        return False

    if os.path.exists(path):
        return True
    else:
        return False


def save_df_as_csv(df, path, filename, comment=None, **kwargs):
    """ Save dataframe to a CSV file.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe to save
    path : str
        path to directory where to save CSV file
    filename : str
        filename of CSV file
    comment : str
        header comment(s); one or more lines starting with '#'
    **kwargs
        additional parameters to `pandas.DataFrame.to_csv`

    Returns
    -------
    str
        path to saved file, else empty str
    """
    if isinstance(df, pd.DataFrame) and len(df) > 0:
        try:
            if not create_dir(path):
                return ''

            destination = os.path.join(path, filename)

            print('Saving ' + os.path.relpath(destination))

            s = '# Generated by lineage v{}, https://github.com/apriha/lineage\n' \
                '# Generated at {} UTC\n'

            s = s.format(__version__, datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'))

            if isinstance(comment, str):
                s += comment

            with open(destination, 'w') as f:
                f.write(s)

            # https://stackoverflow.com/a/29233924/4727627
            with open(destination, 'a') as f:
                df.to_csv(f, na_rep='--', **kwargs)

            return destination
        except Exception as err:
            print(err)
            return ''
    else:
        print('no data to save...')
        return ''



""" Utilities for reading and analyzing SNPs within the `lineage` framework. """

"""
Copyright (C) 2018 Andrew Riha

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from itertools import groupby, count
import gzip
import os
import re
import zipfile

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype

from lineage.ensembl import EnsemblRestClient

class SNPs(object):

    def __init__(self, file, assign_par_snps=True):
        """ Object used to read and parse genotype / raw data files.

        Parameters
        ----------
        file : str
            path to file to load
        assign_par_snps : bool
            assign PAR SNPs to the X and Y chromosomes
        """
        self.snps, self.source = self._read_raw_data(file)
        self.build = None
        self.build_detected = False

        if self.snps is not None:
            self.build = detect_build(self.snps)

            if self.build is None:
                self.build = 37  # assume Build 37 / GRCh37 if not detected
            else:
                self.build_detected = True

            if assign_par_snps:
                self._assign_par_snps()

    @property
    def assembly(self):
        """ Get the assembly of ``SNPs``.

        Returns
        -------
        str
        """
        return get_assembly(self.build)

    @property
    def snp_count(self):
        """ Count of SNPs.

        Returns
        -------
        int
        """
        return get_snp_count(self.snps)

    @property
    def chromosomes(self):
        """ Chromosomes of ``SNPs``.

        Returns
        -------
        list
            list of str chromosomes (e.g., ['1', '2', '3', 'MT'], empty list if no chromosomes
        """
        return get_chromosomes(self.snps)

    @property
    def chromosomes_summary(self):
        """ Summary of the chromosomes of ``SNPs``.

        Returns
        -------
        str
            human-readable listing of chromosomes (e.g., '1-3, MT'), empty str if no chromosomes
        """
        return get_chromosomes_summary(self.snps)

    @property
    def sex(self):
        """ Sex derived from ``SNPs``.

        Returns
        -------
        str
            'Male' or 'Female' if detected, else empty str
        """
        return determine_sex(self.snps)

    def get_summary(self):
        """ Get summary of ``SNPs``.

        Returns
        -------
        dict
            summary info, else None if ``SNPs`` is not valid
        """
        if not self.is_valid():
            return None
        else:
            return {'source': self.source,
                    'assembly': self.assembly,
                    'build': self.build,
                    'build_detected': self.build_detected,
                    'snp_count': self.snp_count,
                    'chromosomes': self.chromosomes_summary,
                    'sex': self.sex}

    def is_valid(self):
        """ Determine if ``SNPs`` is valid.

        ``SNPs`` is valid when the input file has been successfully parsed.

        Returns
        -------
        bool
            True if ``SNPs`` is valid
        """
        if self.snps is None:
            return False
        else:
            return True

    def _read_raw_data(self, file):
        try:
            if not os.path.exists(file):
                print(file + ' does not exist; skipping')
                return None, ''

            # peek into files to determine the data format
            if '.zip' in file:
                with zipfile.ZipFile(file) as z:
                    with z.open(z.namelist()[0], 'r') as f:
                        first_line, comments = self._extract_comments(f, True)
            elif '.gz' in file:
                with gzip.open(file, 'rt') as f:
                    first_line, comments = self._extract_comments(f, False)
            else:
                with open(file, 'r') as f:
                    first_line, comments = self._extract_comments(f, False)

            if '23andMe' in first_line:
                return self._read_23andme(file)
            elif 'Ancestry' in first_line:
                return self._read_ancestry(file)
            elif first_line.startswith('RSID'):
                return self._read_ftdna(file)
            elif 'lineage' in first_line:
                return self._read_lineage_csv(file, comments)
            elif first_line.startswith('rsid'):
                return self._read_generic_csv(file)
            else:
                return None, ''
        except Exception as err:
            print(err)
            return None, ''

    def _extract_comments(self, f, decode):
        line = self._read_line(f, decode)
        first_line = line
        comments = ''

        while line.startswith('#'):
            comments += line
            line = self._read_line(f, decode)

        return first_line, comments

    def _read_line(self, f, decode):
        if decode:
            # https://stackoverflow.com/a/606199
            return f.readline().decode('utf-8')
        else:
            return f.readline()

    @staticmethod
    def _read_23andme(file):
        """ Read and parse 23andMe file.

        https://www.23andme.com

        Parameters
        ----------
        file : str
            path to file

        Returns
        -------
        pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        str
            name of data source
        """
        df = pd.read_csv(file, comment='#', sep='\t', na_values='--',
                               names=['rsid', 'chrom', 'pos', 'genotype'],
                               index_col=0, dtype={'chrom': object})

        return sort_snps(df), '23andMe'

    @staticmethod
    def _read_ftdna(file):
        """ Read and parse Family Tree DNA (FTDNA) file.

        https://www.familytreedna.com

        Parameters
        ----------
        file : str
            path to file

        Returns
        -------
        pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        str
            name of data source
        """
        df = pd.read_csv(file, skiprows=1, na_values='--',
                         names=['rsid', 'chrom', 'pos', 'genotype'],
                         index_col=0, dtype={'chrom': object})

        # remove incongruous data
        df = df.drop(df.loc[df['chrom'] == '0'].index)
        df = df.drop(df.loc[df.index == 'RSID'].index)  # second header for concatenated data

        # if second header existed, pos dtype will be object (should be np.int64)
        df['pos'] = df['pos'].astype(np.int64)

        return sort_snps(df), 'FTDNA'

    @staticmethod
    def _read_ancestry(file):
        """ Read and parse Ancestry.com file.

        http://www.ancestry.com

        Parameters
        ----------
        file : str
            path to file

        Returns
        -------
        pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        str
            name of data source
        """
        df = pd.read_csv(file, comment='#', header=0, sep='\t', na_values=0,
                         names=['rsid', 'chrom', 'pos', 'allele1', 'allele2'],
                         index_col=0, dtype={'chrom': object})

        # create genotype column from allele columns
        df['genotype'] = df['allele1'] + df['allele2']

        # delete allele columns
        # http://stackoverflow.com/a/13485766
        del df['allele1']
        del df['allele2']

        # https://redd.it/5y90un
        df.ix[np.where(df['chrom'] == '23')[0], 'chrom'] = 'X'
        df.ix[np.where(df['chrom'] == '24')[0], 'chrom'] = 'Y'
        df.ix[np.where(df['chrom'] == '25')[0], 'chrom'] = 'PAR'
        df.ix[np.where(df['chrom'] == '26')[0], 'chrom'] = 'MT'

        return sort_snps(df), 'AncestryDNA'

    @staticmethod
    def _read_lineage_csv(file, comments):
        """ Read and parse CSV file generated by lineage.

        Parameters
        ----------
        file : str
            path to file
        comments : str
            comments at beginning of file

        Returns
        -------
        pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        str
            name of data source(s)
        """
        source = ''
        for comment in comments.split('\n'):
            if 'Source(s):' in comment:
                source = comment.split('Source(s):')[1].strip()
                break

        df = pd.read_csv(file, comment='#', header=0, na_values='--',
                         names=['rsid', 'chrom', 'pos', 'genotype'],
                         index_col=0, dtype={'chrom': object, 'pos': np.int64})

        return sort_snps(df), source

    @staticmethod
    def _read_generic_csv(file):
        """ Read and parse generic CSV file.

        Notes
        -----
        Assumes columns are 'rsid', 'chrom' / 'chromosome', 'pos' / 'position', and 'genotype';
        values are comma separated; unreported genotypes are indicated by '--'; and one header row
        precedes data. For example:

            rsid,chromosome,position,genotype
            rs1,1,1,AA
            rs2,1,2,CC
            rs3,1,3,--

        Parameters
        ----------
        file : str
            path to file

        Returns
        -------
        pandas.DataFrame
            individual's genetic data normalized for use with `lineage`
        str
            name of data source
        """
        df = pd.read_csv(file, skiprows=1, na_values='--',
                         names=['rsid', 'chrom', 'pos', 'genotype'],
                         index_col=0, dtype={'chrom': object, 'pos': np.int64})

        return sort_snps(df), 'generic'

    def _assign_par_snps(self):
        """ Assign PAR SNPs to the X or Y chromosome using SNP position.

        References
        -----
        ..[1] National Center for Biotechnology Information, Variation Services, RefSNP,
          https://api.ncbi.nlm.nih.gov/variation/v0/
        ..[2] Yates et. al. (doi:10.1093/bioinformatics/btu613),
          http://europepmc.org/search/?query=DOI:10.1093/bioinformatics/btu613
        ..[3] Zerbino et. al. (doi.org/10.1093/nar/gkx1098), https://doi.org/10.1093/nar/gkx1098
        ..[4] Sherry ST, Ward MH, Kholodov M, Baker J, Phan L, Smigielski EM, Sirotkin K.
          dbSNP: the NCBI database of genetic variation. Nucleic Acids Res. 2001 Jan 1;
          29(1):308-11.
        ..[5] Database of Single Nucleotide Polymorphisms (dbSNP). Bethesda (MD): National Center
          for Biotechnology Information, National Library of Medicine. dbSNP accession: rs28736870,
          rs113313554, and rs758419898 (dbSNP Build ID: 151). Available from:
          http://www.ncbi.nlm.nih.gov/SNP/
        """
        rest_client = EnsemblRestClient(server='https://api.ncbi.nlm.nih.gov')
        for rsid in self.snps.loc[self.snps['chrom'] == 'PAR'].index.values:
            if 'rs' in rsid:
                try:
                    id = rsid.split('rs')[1]
                    response = rest_client.perform_rest_action('/variation/v0/beta/refsnp/' + id)

                    if response is not None:
                        for item in response['primary_snapshot_data']['placements_with_allele']:
                            if 'NC_000023' in item['seq_id']:
                                assigned = self._assign_snp(rsid, item['alleles'], 'X')
                            elif 'NC_000024' in item['seq_id']:
                                assigned = self._assign_snp(rsid, item['alleles'], 'Y')
                            else:
                                assigned = False

                            if assigned:
                                if not self.build_detected:
                                    self.build = self._extract_build(item)
                                    self.build_detected = True
                                continue

                except Exception as err:
                    print(err)

    def _assign_snp(self, rsid, alleles, chrom):
        for allele in alleles:
            allele_pos = allele['allele']['spdi']['position']
            # ref SNP positions seem to be 0-based...
            if allele_pos == self.snps.loc[rsid].pos - 1:
                self.snps.loc[rsid, 'chrom'] = chrom
                return True
        return False

    def _extract_build(self, item):
        assembly_name = item['placement_annot']['seq_id_traits_by_assembly'][0]['assembly_name']
        assembly_name = assembly_name.split('.')[0]
        return int(assembly_name[-2:])

def detect_build(snps):
    """ Detect build of SNPs.

    Use the coordinates of common SNPs to identify the build / assembly of a genotype file
    that is being loaded.

    Notes
    -----
    rs3094315 : plus strand in 36, 37, and 38
    rs11928389 : plus strand in 36, minus strand in 37 and 38
    rs2500347 : plus strand in 36 and 37, minus strand in 38
    rs964481 : plus strand in 36, 37, and 38

    Parameters
    ----------
    snps : pandas.DataFrame
        SNPs to add

    Returns
    -------
    int
        detected build of SNPs, else None

    References
    ----------
    ..[1] Yates et. al. (doi:10.1093/bioinformatics/btu613),
      http://europepmc.org/search/?query=DOI:10.1093/bioinformatics/btu613
    ..[2] Zerbino et. al. (doi.org/10.1093/nar/gkx1098), https://doi.org/10.1093/nar/gkx1098
    ..[3] Sherry ST, Ward MH, Kholodov M, Baker J, Phan L, Smigielski EM, Sirotkin K.
      dbSNP: the NCBI database of genetic variation. Nucleic Acids Res. 2001 Jan 1;29(1):308-11.
    ..[4] Database of Single Nucleotide Polymorphisms (dbSNP). Bethesda (MD): National Center
      for Biotechnology Information, National Library of Medicine. dbSNP accession: rs3094315,
      rs11928389, rs2500347, and rs964481 (dbSNP Build ID: 151). Available from:
      http://www.ncbi.nlm.nih.gov/SNP/
    """
    def lookup_build_with_snp_pos(pos, s):
        try:
            return s.loc[s == pos].index[0]
        except:
            return None

    build = None

    rsids = ['rs3094315', 'rs11928389', 'rs2500347', 'rs964481']
    df = pd.DataFrame({36: [742429, 50908372, 143649677, 27566744],
                       37: [752566, 50927009, 144938320, 27656823],
                       38: [817186, 50889578, 148946169, 27638706]}, index=rsids)

    for rsid in rsids:
        if rsid in snps.index:
            build = lookup_build_with_snp_pos(snps.loc[rsid].pos, df.loc[rsid])

        if build is not None:
            break

    return build


def get_assembly(build):
    """ Get the assembly of a build.

    Parameters
    ----------
    build : int {36, 37, 38}

    Returns
    -------
    str
        empty str if `build` is None
    """

    if build is None:
        return ''
    elif build == 36:
        return 'NCBI36'
    elif build == 38:
        return 'GRCh38'
    else:
        return 'GRCh37'


def get_snp_count(snps):
    """ Count of SNPs.

    Parameters
    ----------
    snps : pandas.DataFrame

    Returns
    -------
    int
    """

    if snps is not None:
        return len(snps)
    else:
        return 0


def get_chromosomes(snps):
    """ Get the chromosomes of SNPs.

    Parameters
    ----------
    snps : pandas.DataFrame

    Returns
    -------
    list
        list of str chromosomes (e.g., ['1', '2', '3', 'MT'], empty list if no chromosomes
    """

    if isinstance(snps, pd.DataFrame):
        return list(pd.unique(snps['chrom']))
    else:
        return []


def get_chromosomes_summary(snps):
    """ Summary of the chromosomes of SNPs.

    Parameters
    ----------
    snps : pandas.DataFrame

    Returns
    -------
    str
        human-readable listing of chromosomes (e.g., '1-3, MT'), empty str if no chromosomes
    """

    if isinstance(snps, pd.DataFrame):
        chroms = list(pd.unique(snps['chrom']))

        int_chroms = [int(chrom) for chrom in chroms if chrom.isdigit()]
        str_chroms = [chrom for chrom in chroms if not chrom.isdigit()]

        # https://codereview.stackexchange.com/a/5202
        def as_range(iterable):
            l = list(iterable)
            if len(l) > 1:
                return '{0}-{1}'.format(l[0], l[-1])
            else:
                return '{0}'.format(l[0])

        # create str representations
        int_chroms = ', '.join(as_range(g) for _, g in
                               groupby(int_chroms, key=lambda n, c=count(): n - next(c)))
        str_chroms = ', '.join(str_chroms)

        if int_chroms != '' and str_chroms != '':
            int_chroms += ', '

        return int_chroms + str_chroms
    else:
        return ''

def determine_sex(snps, y_snps_not_null_threshold=0.1, heterozygous_x_snps_threshold=0.01):
    """ Determine sex from SNPs using thresholds.

    Parameters
    ----------
    snps : pandas.DataFrame
    y_snps_not_null_threshold : float
        percentage Y SNPs that are not null; above this threshold, Male is determined
    heterozygous_x_snps_threshold : float
        percentage heterozygous X SNPs; above this threshold, Female is determined

    Returns
    -------
    str
        'Male' or 'Female' if detected, else empty str
    """

    if isinstance(snps, pd.DataFrame):
        y_snps = len(snps.loc[(snps['chrom'] == 'Y')])

        if y_snps > 0:
            y_snps_not_null = len(snps.loc[(snps['chrom'] == 'Y') & (snps['genotype'].notnull())])

            if y_snps_not_null / y_snps > y_snps_not_null_threshold:
                return 'Male'
            else:
                return 'Female'

        x_snps = len(snps.loc[snps['chrom'] == 'X'])

        if x_snps == 0:
            return ''

        heterozygous_x_snps = len(snps.loc[(snps['chrom'] == 'X') &
                                           (snps['genotype'].notnull()) &
                                           (snps['genotype'].str[0] != snps['genotype'].str[1])])

        if heterozygous_x_snps / x_snps > heterozygous_x_snps_threshold:
            return 'Female'
        else:
            return 'Male'
    else:
        return ''

def sort_snps(snps):
    """ Sort SNPs based on ordered chromosome list and position. """

    sorted_list = sorted(snps['chrom'].unique(), key=_natural_sort_key)

    # move PAR and MT to the end of the dataframe
    if 'PAR' in sorted_list:
        sorted_list.remove('PAR')
        sorted_list.append('PAR')

    if 'MT' in sorted_list:
        sorted_list.remove('MT')
        sorted_list.append('MT')

    # convert chrom column to category for sorting
    # https://stackoverflow.com/a/26707444
    snps['chrom'] = \
        snps['chrom'].astype(CategoricalDtype(categories=sorted_list, ordered=True))

    # sort based on ordered chromosome list and position
    snps = snps.sort_values(['chrom', 'pos'])

    # convert chromosome back to object
    snps['chrom'] = snps['chrom'].astype(object)

    return snps

# https://stackoverflow.com/a/16090640
def _natural_sort_key(s, natural_sort_re=re.compile('([0-9]+)')):
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split(natural_sort_re, s)]



""" Class for downloading and loading required external resources.

`lineage` uses tables and data from UCSC's Genome Browser:

* http://genome.ucsc.edu/
* http://genome.ucsc.edu/cgi-bin/hgTables

References
----------
..[1] Karolchik D, Hinrichs AS, Furey TS, Roskin KM, Sugnet CW, Haussler D, Kent WJ.
  The UCSC Table Browser data retrieval tool. Nucleic Acids Res. 2004 Jan
  1;32(Database issue):D493-6. PubMed PMID: 14681465; PubMed Central PMCID:
  PMC308837. https://www.ncbi.nlm.nih.gov/pubmed/14681465
..[2] Tyner C, Barber GP, Casper J, Clawson H, Diekhans M, Eisenhart C, Fischer CM,
  Gibson D, Gonzalez JN, Guruvadoo L, Haeussler M, Heitner S, Hinrichs AS,
  Karolchik D, Lee BT, Lee CM, Nejad P, Raney BJ, Rosenbloom KR, Speir ML,
  Villarreal C, Vivian J, Zweig AS, Haussler D, Kuhn RM, Kent WJ. The UCSC Genome
  Browser database: 2017 update. Nucleic Acids Res. 2017 Jan 4;45(D1):D626-D634.
  doi: 10.1093/nar/gkw1134. Epub 2016 Nov 29. PubMed PMID: 27899642; PubMed Central
  PMCID: PMC5210591. https://www.ncbi.nlm.nih.gov/pubmed/27899642
..[3] International Human Genome Sequencing Consortium. Initial sequencing and
  analysis of the human genome. Nature. 2001 Feb 15;409(6822):860-921.
  http://dx.doi.org/10.1038/35057062
..[4] hg19 (GRCh37): Hiram Clawson, Brooke Rhead, Pauline Fujita, Ann Zweig, Katrina
  Learned, Donna Karolchik and Robert Kuhn, https://genome.ucsc.edu/cgi-bin/hgGateway?db=hg19
..[5] Yates et. al. (doi:10.1093/bioinformatics/btu613),
  http://europepmc.org/search/?query=DOI:10.1093/bioinformatics/btu613
..[6] Zerbino et. al. (doi.org/10.1093/nar/gkx1098), https://doi.org/10.1093/nar/gkx1098

"""

"""
Copyright (C) 2017 Andrew Riha

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import gzip
import itertools
import json
import os
import tarfile
import tempfile
import urllib.error
import urllib.request
import zlib

import pandas as pd

import lineage


class Resources(object):
    """ Object used to manage resources required by `lineage`. """

    def __init__(self, resources_dir='resources', ensembl_rest_client=None):
        """ Initialize a ``Resources`` object.

        Parameters
        ----------
        resources_dir : str
            name / path of resources directory
        """
        self._resources_dir = os.path.abspath(resources_dir)
        self._genetic_map_HapMapII_GRCh37 = None
        self._cytoBand_hg19 = None
        self._knownGene_hg19 = None
        self._kgXref_hg19 = None
        self._ensembl_rest_client = ensembl_rest_client

    def get_genetic_map_HapMapII_GRCh37(self):
        """ Get International HapMap Consortium HapMap Phase II genetic map for Build 37.

        Returns
        -------
        dict
            dict of pandas.DataFrame HapMapII genetic maps if loading was successful, else None
        """
        if self._genetic_map_HapMapII_GRCh37 is None:
            self._genetic_map_HapMapII_GRCh37 = \
                self._load_genetic_map(self._get_path_genetic_map_HapMapII_GRCh37())

        return self._genetic_map_HapMapII_GRCh37

    def get_cytoBand_hg19(self):
        """ Get UCSC cytoBand table for Build 37.

        Returns
        -------
        pandas.DataFrame
            cytoBand table if loading was successful, else None
        """
        if self._cytoBand_hg19 is None:
            self._cytoBand_hg19 = self._load_cytoBand(self._get_path_cytoBand_hg19())

        return self._cytoBand_hg19

    def get_knownGene_hg19(self):
        """ Get UCSC knownGene table for Build 37.

        Returns
        -------
        pandas.DataFrame
            knownGene table if loading was successful, else None
        """
        if self._knownGene_hg19 is None:
            self._knownGene_hg19 = self._load_knownGene(self._get_path_knownGene_hg19())

        return self._knownGene_hg19

    def get_kgXref_hg19(self):
        """ Get UCSC kgXref table for Build 37.

        Returns
        -------
        pandas.DataFrame
            kgXref table if loading was successful, else None
        """
        if self._kgXref_hg19 is None:
            self._kgXref_hg19 = self._load_kgXref(self._get_path_kgXref_hg19())

        return self._kgXref_hg19

    def get_assembly_mapping_data(self, source_assembly, target_assembly):
        """ Get assembly mapping data.

        Parameters
        ----------
        source_assembly : {'NCBI36', 'GRCh37', 'GRCh38'}
            assembly to remap from
        target_assembly : {'NCBI36', 'GRCh37', 'GRCh38'}
            assembly to remap to

        Returns
        -------
        dict
            dict of json assembly mapping data if loading was successful, else None
        """
        return self._load_assembly_mapping_data(
            self._get_path_assembly_mapping_data(source_assembly, target_assembly))

    def download_example_datasets(self):
        """ Download example datasets from `openSNP <https://opensnp.org>`_.

        Per openSNP, "the data is donated into the public domain using `CC0 1.0
        <http://creativecommons.org/publicdomain/zero/1.0/>`_."

        Returns
        -------
        paths : list of str or None
            paths to example datasets

        References
        ----------
        ..[1] Greshake B, Bayer PE, Rausch H, Reda J (2014), "openSNP-A Crowdsourced Web Resource
          for Personal Genomics," PLOS ONE, 9(3): e89204,
          https://doi.org/10.1371/journal.pone.0089204
        """
        paths = []
        paths.append(self._download_file('https://opensnp.org/data/662.23andme.304',
                                         '662.23andme.304.txt.gz', compress=True))
        paths.append(self._download_file('https://opensnp.org/data/662.23andme.340',
                                         '662.23andme.340.txt.gz', compress=True))
        paths.append(self._download_file('https://opensnp.org/data/662.ftdna-illumina.341',
                                         '662.ftdna-illumina.341.csv.gz', compress=True))
        paths.append(self._download_file('https://opensnp.org/data/663.23andme.305',
                                         '663.23andme.305.txt.gz', compress=True))

        # these two files consist of concatenated gzip files and therefore need special handling
        paths.append(self._download_file('https://opensnp.org/data/4583.ftdna-illumina.3482',
                                         '4583.ftdna-illumina.3482.csv.gz'))
        paths.append(self._download_file('https://opensnp.org/data/4584.ftdna-illumina.3483',
                                         '4584.ftdna-illumina.3483.csv.gz'))

        try:
            for gzip_path in paths[-2:]:
                # https://stackoverflow.com/q/4928560
                # https://stackoverflow.com/a/37042747
                with open(gzip_path, 'rb') as f:
                    decompressor = zlib.decompressobj(31)

                    # decompress data from first concatenated gzip file
                    data = decompressor.decompress(f.read())

                    if len(decompressor.unused_data) > 0:
                        # decompress data from second concatenated gzip file, if any
                        additional_data = zlib.decompress(decompressor.unused_data, 31)
                        data += additional_data[33:]  # skip over second header

                # recompress data
                with gzip.open(gzip_path, 'wb') as f:
                    f.write(data)
        except Exception as err:
            print(err)

        return paths

    def get_all_resources(self):
        """ Get / download all resources used throughout `lineage`.

        Returns
        -------
        dict
            dict of resources
        """
        resources = {}
        resources['genetic_map_HapMapII_GRCh37'] = self.get_genetic_map_HapMapII_GRCh37()
        resources['cytoBand_hg19'] = self.get_cytoBand_hg19()
        resources['knownGene_hg19'] = self.get_knownGene_hg19()
        resources['kgXref_hg19'] = self.get_kgXref_hg19()
        for source, target in itertools.permutations(['NCBI36', 'GRCh37', 'GRCh38'], 2):
            resources[source + '_' + target] = self.get_assembly_mapping_data(source, target)
        return resources

    @staticmethod
    def _load_genetic_map(filename):
        """ Load genetic map (e.g. HapMapII).

        Parameters
        ----------
        filename : str
            path to compressed archive with genetic map data

        Returns
        -------
        genetic_map : dict
            dict of pandas.DataFrame genetic maps if loading was successful, else None

        Notes
        -----
        Keys of returned dict are chromosomes and values are the corresponding genetic map.
        """
        try:
            genetic_map = {}

            with tarfile.open(filename, 'r') as tar:
                # http://stackoverflow.com/a/2018576
                for member in tar.getmembers():
                    if 'genetic_map' in member.name:
                        df = pd.read_csv(tar.extractfile(member), sep='\t')
                        df = df.rename(columns={'Position(bp)': 'pos',
                                                'Rate(cM/Mb)': 'rate',
                                                'Map(cM)': 'map'})
                        del df['Chromosome']
                        start_pos = member.name.index('chr') + 3
                        end_pos = member.name.index('.')
                        genetic_map[member.name[start_pos:end_pos]] = df

            # X chrom consists of X PAR regions and X non-PAR region
            genetic_map['X'] = pd.concat([genetic_map['X_par1'], genetic_map['X'],
                                          genetic_map['X_par2']])
            del genetic_map['X_par1']
            del genetic_map['X_par2']

            return genetic_map
        except Exception as err:
            print(err)
            return None

    @staticmethod
    def _load_assembly_mapping_data(filename):
        """ Load assembly mapping data.

        Parameters
        ----------
        filename : str
            path to compressed archive with assembly mapping data

        Returns
        -------
        assembly_mapping_data : dict
            dict of assembly maps if loading was successful, else None

        Notes
        -----
        Keys of returned dict are chromosomes and values are the corresponding assembly map.
        """
        try:
            assembly_mapping_data = {}

            with tarfile.open(filename, 'r') as tar:
                # http://stackoverflow.com/a/2018576
                for member in tar.getmembers():
                    if '.json' in member.name:
                        with tar.extractfile(member) as tar_file:
                            tar_bytes = tar_file.read()
                        # https://stackoverflow.com/a/42683509/4727627
                        assembly_mapping_data[member.name.split('.')[0]] =\
                            json.loads(tar_bytes.decode('utf-8'))

            return assembly_mapping_data
        except Exception as err:
            print(err)
            return None

    @staticmethod
    def _load_cytoBand(filename):
        """ Load UCSC cytoBand table.

        Parameters
        ----------
        filename : str
            path to cytoBand file

        Returns
        -------
        df : pandas.DataFrame
            cytoBand table if loading was successful, else None

        References
        ----------
        ..[1] Ryan Dale, GitHub Gist,
          https://gist.github.com/daler/c98fc410282d7570efc3#file-ideograms-py
        """
        try:
            # adapted from chromosome plotting code (see [1]_)
            df = pd.read_table(filename, names=['chrom', 'start', 'end', 'name', 'gie_stain'])
            df['chrom'] = df['chrom'].str[3:]
            return df
        except Exception as err:
            print(err)
            return None

    @staticmethod
    def _load_knownGene(filename):
        """ Load UCSC knownGene table.

        Parameters
        ----------
        filename : str
            path to knownGene file

        Returns
        -------
        df : pandas.DataFrame
            knownGene table if loading was successful, else None
        """
        try:
            df = pd.read_table(filename, names=['name', 'chrom', 'strand', 'txStart', 'txEnd',
                                                'cdsStart', 'cdsEnd', 'exonCount', 'exonStarts',
                                                'exonEnds', 'proteinID', 'alignID'], index_col=0)
            df['chrom'] = df['chrom'].str[3:]
            return df
        except Exception as err:
            print(err)
            return None

    @staticmethod
    def _load_kgXref(filename):
        """ Load UCSC kgXref table.

        Parameters
        ----------
        filename : str
            path to kgXref file

        Returns
        -------
        df : pandas.DataFrame
            kgXref table if loading was successful, else None
        """
        try:
            df = pd.read_table(filename, names=['kgID', 'mRNA', 'spID', 'spDisplayID',
                                                'geneSymbol', 'refseq', 'protAcc',
                                                'description', 'rfamAcc', 'tRnaName'], index_col=0,
                               dtype=object)
            return df
        except Exception as err:
            print(err)
            return None

    def _get_path_cytoBand_hg19(self):
        """ Get local path to cytoBand file for hg19 / GRCh37 from UCSC, downloading if necessary.

        Returns
        -------
        str
            path to cytoBand_hg19.txt.gz
        """
        return self._download_file(
            'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/database/cytoBand.txt.gz',
            'cytoBand_hg19.txt.gz')

    def _get_path_genetic_map_HapMapII_GRCh37(self):
        """ Get local path to HapMap Phase II genetic map for hg19 / GRCh37 (HapMapII),
        downloading if necessary.

        Returns
        -------
        str
            path to genetic_map_HapMapII_GRCh37.tar.gz

        References
        ----------
        ..[1] "The International HapMap Consortium (2007).  A second generation human haplotype
          map of over 3.1 million SNPs.  Nature 449: 851-861."
        ..[2] "The map was generated by lifting the HapMap Phase II genetic map from build 35 to
          GRCh37. The original map was generated using LDhat as described in the 2007 HapMap
          paper (Nature, 18th Sept 2007). The conversion from b35 to GRCh37 was achieved using
          the UCSC liftOver tool. Adam Auton, 08/12/2010"
        """
        return self._download_file(
            'ftp://ftp.ncbi.nlm.nih.gov/hapmap/recombination/2011-01_phaseII_B37/'
            'genetic_map_HapMapII_GRCh37.tar.gz', 'genetic_map_HapMapII_GRCh37.tar.gz')

    def _get_path_knownGene_hg19(self):
        """ Get local path to knownGene file for hg19 / GRCh37 from UCSC, downloading if necessary.

        Returns
        -------
        str
            path to knownGene_hg19.txt.gz
        """
        return self._download_file(
            'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/database/knownGene.txt.gz',
            'knownGene_hg19.txt.gz')

    def _get_path_kgXref_hg19(self):
        """ Get local path to kgXref file for hg19 / GRCh37 from UCSC, downloading if necessary.

        Returns
        -------
        str
            path to kgXref_hg19.txt.gz
        """
        return self._download_file(
            'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg19/database/kgXref.txt.gz',
            'kgXref_hg19.txt.gz')

    def _get_path_assembly_mapping_data(self, source_assembly, target_assembly, retries=10):
        """ Get local path to assembly mapping data, downloading if necessary.

        Parameters
        ----------
        source_assembly : {'NCBI36', 'GRCh37', 'GRCh38'}
            assembly to remap from
        target_assembly : {'NCBI36', 'GRCh37', 'GRCh38'}
            assembly to remap to
        retries : int
            number of retries per chromosome to download assembly mapping data

        Returns
        -------
        str
            path to <source_assembly>_<target_assembly>.tar.gz

        References
        ----------
        ..[1] Ensembl, Assembly Information Endpoint,
          https://rest.ensembl.org/documentation/info/assembly_info
        ..[2] Ensembl, Assembly Map Endpoint,
          http://rest.ensembl.org/documentation/info/assembly_map

        """

        if not lineage.create_dir(self._resources_dir):
            return None

        chroms = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14',
                  '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', 'MT']

        assembly_mapping_data = source_assembly + '_' + target_assembly
        destination = os.path.join(self._resources_dir, assembly_mapping_data + '.tar.gz')

        if not os.path.exists(destination) or not self._all_chroms_in_tar(chroms, destination):
            print('Downloading {}'.format(os.path.relpath(destination)))

            try:
                with tarfile.open(destination, 'w:gz') as out_tar:
                    for chrom in chroms:
                        file = chrom + '.json'

                        map_endpoint = '/map/human/' + source_assembly + '/' + chrom + '/' + \
                                       target_assembly + '?'

                        # get assembly mapping data
                        response = None
                        retry = 0
                        while response is None and retry < retries:
                            response = self._ensembl_rest_client.perform_rest_action(map_endpoint)
                            retry += 1

                        if response is not None:
                            # open temp file, save json response to file, close temp file
                            with tempfile.NamedTemporaryFile(delete=False, mode='w') as f:
                                json.dump(response, f)

                            # add temp file to archive
                            out_tar.add(f.name, arcname=file)

                            # remove temp file
                            os.remove(f.name)
            except Exception as err:
                print(err)
                return None

        return destination

    def _all_chroms_in_tar(self, chroms, filename):
        try:
            with tarfile.open(filename, 'r') as tar:
                members = tar.getnames()

            for chrom in chroms:
                if chrom + '.json' not in members:
                    return False
        except Exception as err:
            print(err)
            return False

        return True

    def _download_file(self, url, filename, compress=False, timeout=30):
        """ Download a file to the resources folder.

        Download data from `url`, save as `filename`, and optionally compress with gzip.

        Parameters
        ----------
        url : str
            URL to download data from
        filename : str
            name of file to save; if compress, ensure '.gz' is appended
        compress : bool
            compress with gzip
        timeout : int
            seconds for timeout of download request

        Returns
        -------
        str
            path to downloaded file, None if error
        """
        if not lineage.create_dir(self._resources_dir):
            return None

        if compress and filename[-3:] != '.gz':
            filename += '.gz'

        destination = os.path.join(self._resources_dir, filename)

        if not os.path.exists(destination):
            try:
                if compress:
                    open_func = gzip.open
                else:
                    open_func = open

                # get file if it hasn't already been downloaded
                # http://stackoverflow.com/a/7244263
                with urllib.request.urlopen(url, timeout=timeout) as response, \
                        open_func(destination, 'wb') as f:
                    self._print_download_msg(destination)
                    data = response.read()  # a `bytes` object
                    f.write(data)
            except urllib.error.URLError as err:
                print(err)
                destination = None
                # try HTTP if an FTP error occurred
                if 'ftp://' in url:
                    destination = self._download_file(url.replace('ftp://', 'http://'),
                                                      filename, compress=compress, timeout=timeout)
            except Exception as err:
                print(err)
                return None

        return destination

    @staticmethod
    def _print_download_msg(path):
        """ Print download message.

        Parameters
        ----------
        path : str
            path to file being downloaded
        """
        print('Downloading ' + os.path.relpath(path))
